[org 0]
[bits 16]

start:
    mov ax, cs          ; CS 0x1000 ִ.
    mov ds, ax
    xor ax, ax
    mov ss, ax;

    cli

    lgdt[gdtr]

    mov eax, cr0
    or  eax, 0x00000001
    mov cr0, eax

    jmp $+2
    nop
    nop

    db  0x66
    db  0x67
    db  0xEA
    dd  PM_Start
    dw  SysCodeSelector

;-------------------------------------------------------------------------
;            protected mode Դϴ.
;-------------------------------------------------------------------------
[bits 32]

PM_Start:
    mov bx, SysDataSelector
    mov ds, bx
    mov es, bx
    mov fs, bx
    mov gs, bx
    mov ss, bx

    xor eax, eax
    mov ax, VideoSelector
    mov es, ax
    mov edi, 80*2*10 + 2*10         ; ȭ(x, y) = (10, 10)
    lea esi, [ds:msgPMode]
    call    printf

    jmp $


;-------------------------------------------------------------------------
;           Sub Routines
;-------------------------------------------------------------------------
printf:
    push    eax

printf_loop:
    or  al, al
    jz  printf_end
    mov al, byte[esi]
    mov byte[es:edi], al
    inc edi
    mov byte[es:edi], 0x06
    inc esi
    inc edi
    jmp printf_loop

printf_end:
    pop eax
    ret

msgPMode    db  "Three-Run.. We are in Protected Mode!!!", 0

;-------------------------------------------------------------------------
;           GDT(Global Descriptor Table)
;               BaseAddress, Descriptor size, Atribute
;-------------------------------------------------------------------------
gdtr:
    dw  gdt_end - gdt -1    ; Limit of GDT  == sizeof(gdtr)
    dd  gdt + 0x10000       ; base address of GDT
    16 + 32 = 48bit

;-------------------------------------------------------------------------
;              Descriptor
;   +---------------+---------------+
;   |   Limit(15 ~ 0 bit)           |
;   +---------------+---------------+
;   |   Base Address(15 ~ 0 bit)    |
;   +---------------+---------------+
;   |P |DPL |S|Type | Base 23 ~ 16  |
;   +---------------+---------------+
;   |Base 31 ~24    |G|D|0|A|limit  |
;   +---------------+---------------+
;-------------------------------------------------------------------------
gdt:
; NULL Descriptor
    dw  0                   ; limit 0 ~ 15 bit
    dw  0                   ; base address   byte
    db  0                   ; base address 16 ~ 32
    db  0                   ; type
    db  0                   ; limit 15 ~ 19 bit, Flag
    db  0                   ; base address 24 ~ 31 bit

; Code segment Descriptor
SysCodeSelector equ 0x08    ; 0000 1000 ==> 00001 000
    dw  0xFFFF              ; limit 0 ~ 15bit
    dw  0x0000              ; base address   byte
    db  0x01                ; base address 16 ~ 32
    db  0x9A                ; type(P:1, DPL:0, Code, non-conforming, readable
    db  0xCF                ; limit 15 ~ 19 bit, Flag(G:0, D:1, limit 0xF)
    db  0x00                ; base address 24 ~ 31 bit

; Data segment Descriptor
SysDataSelector equ 0x10    ; 0001 0000 ==> 00010 000
    dw  0xFFFF              ; limit 0 ~ 15bit
    dw  0x0000              ; base address   byte
    db  0x01                ; base address 16 ~ 32
    db  0x92                ; type(P:1, DPL:0, Data, expand-up, writable
    db  0xCF                ; limit 15 ~ 19 bit, Flag(G:0, D:1, limit 0xF)
    db  0x00                ; base address 24 ~ 31 bit

; Video segment Descriptor
VideoSelector   equ 0x18    ; 0001 1000 ==> 00011 000
    dw  0xFFFF              ; limit 0 ~ 15bit
    dw  0x8000              ; base address   byte
    db  0x0B                ; base address 16 ~ 32
    db  0x92                ; type(P:1, DPL:0, Data, expand-up, writable
    db  0x40                ; limit 15 ~ 19 bit, Flag(G:0, D:1, limit 0x0)
    db  0x00                ; base address 24 ~ 31 bit

gdt_end:
